#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>    // ioctl, _IOR, _IOWR
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include "../include/OSD_flash.h"

#define DBGTRACE(fmt, ...)			fprintf(stderr, "\e[32m" fmt "\e[0m", ## __VA_ARGS__)
//#define DBGTRACE					//

#define CHECK_SUM_R  0
#define CHECK_SUM_W  1

static uint32_t WriteCheckSum;
static uint32_t WriteEvenCheckSum;
static uint32_t WriteOddCheckSum;
static uint32_t ReadCheckSum;
static uint32_t ReadEvenCheckSum;
static uint32_t ReadOddCheckSum;

/* -------------------------------------------------------------------------------- */
/*   ミリ秒待ち                                                                     */
/*                                                                                  */
/*   IN : msec = ミリ秒                                                             */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void msleep( uint32_t msec )
{
	struct timespec		ts ;


	ts.tv_sec  = 0 ;
	ts.tv_nsec = msec * 1000 * 1000 ;
	nanosleep(&ts, NULL) ;
}

/* -------------------------------------------------------------------------------- */
/*   ファイル読み込み                                                               */
/*                                                                                  */
/*   IN : p_fileName = ファイル名                                                   */
/*        pp_data    = 読み込みデータの格納アドレスのアドレス                       */
/*   OUT: -1         = 異常終了                                                     */
/*        上記以外ならデータサイズ                                                  */
/* -------------------------------------------------------------------------------- */
static int readFile(const char *p_fileName, void **pp_data)
{
	struct stat		fileStat ;
	void			*p_data ;
	int				fd, result ;

	/* 開く */
	fd = open(&p_fileName[0], O_RDONLY) ;
	if( fd == -1 ) {
		return -1 ;
	}

	p_data = NULL ;

	do {
		/* ファイルサイズ取得 */
		result = fstat(fd, &fileStat) ;
		if( result ) {
			break ;
		}
		if( !fileStat.st_size ) {
			result = -1 ;
			break ;
		}

		/* メモリー確保 */
		p_data = malloc( fileStat.st_size ) ;
		if( p_data == NULL ) {
			result = -1 ;
			break ;
		}

		/* ファイル読み込み */
		if( read(fd, p_data, fileStat.st_size) != fileStat.st_size ) {
			result = -1 ;
			break ;
		}

		/* 正常終了 */
		result = (int )fileStat.st_size ;
	} while( 0 ) ;

	close( fd ) ;

	if( result < 0 ) {
		if( p_data != NULL ) {
			free( p_data ) ;
		}
		return -1 ;
	}

	*pp_data = p_data ;

	return result ;
}

/* -------------------------------------------------------------------------------- */
/*   OSD-Flash Check Sum 値チェック                                                 */
/*                                                                                  */
/*   OUT:  0    == 正常                                                             */
/*        -1    == 異常                                                             */
/* -------------------------------------------------------------------------------- */
static int check_CheckSum( void )
{
	int ret = -1;

	if( WriteCheckSum == ReadCheckSum ) {
		DBGTRACE("Check Sum OK : Write = 0x%08X Read = 0x%08X\n", WriteCheckSum, ReadCheckSum);
		ret = 0;
	} else {
		DBGTRACE("Check Sum NG : Write = 0x%08X Read = 0x%08X\n", WriteCheckSum, ReadCheckSum);
	}

	return ret;
}

/* -------------------------------------------------------------------------------- */
/*   OSD-Flash Check Sum 計算                                                       */
/*                                                                                  */
/*   IN : mode   = R/Wモード                                                        */
/*   IN : p_data = データ                                                           */
/*        size   = データサイズ                                                     */
/*   OUT:  チェックサム値                                                           */
/* -------------------------------------------------------------------------------- */
static uint32_t calc_CheckSum(uint8_t mode, const uint8_t *p_data, uint32_t size)
{
	uint32_t i;
	uint32_t even, odd;
	uint32_t sum_e, sum_o;
	uint32_t value = 0;


	for( i=0; i<size; i++ ) {

		if( mode == CHECK_SUM_R ) {

			if( (i % 2) == 0 ) {
				ReadEvenCheckSum += p_data[i];
			} else {
				ReadOddCheckSum  += p_data[i];
			}

			sum_e = ReadEvenCheckSum ^ 0xFFFFFFFF;
			sum_o = ReadOddCheckSum  ^ 0xFFFFFFFF;
			value = (sum_e + 1) + (sum_o + 1);
		} else {

			if( (i % 2) == 0 ) {
				WriteEvenCheckSum += p_data[i];
			} else {
				WriteOddCheckSum  += p_data[i];
			}

			sum_e = WriteEvenCheckSum ^ 0xFFFFFFFF;
			sum_o = WriteOddCheckSum  ^ 0xFFFFFFFF;
			value = (sum_e + 1) + (sum_o + 1);
		}
	}

	return value;
}

/* -------------------------------------------------------------------------------- */
/*   OSD-Flash 読み込み                                                             */
/*                                                                                  */
/*   IN : fd     = ファイルディスクリプタ                                           */
/*        size   = データサイズ                                                     */
/*   OUT:  0    == 正常終了                                                         */
/*        -1    == 異常終了                                                         */
/* -------------------------------------------------------------------------------- */
static int flash_read(int fd, uint32_t size)
{
	struct timeval s, e;
	uint32_t readSize = FLASH_RW_DATA_LEN;
	uint32_t check_sum = 0;
	int      result,cnt = 0;
	uint8_t  data[FLASH_RW_DATA_LEN];

	gettimeofday(&s, NULL);
	do {
		if( readSize > size ) {
			readSize = size;
		}

		/* データ読み込み */
		result = read(fd, data, readSize);
		if( result ) {
			DBGTRACE("%s(%d) : %d\n", __func__, __LINE__, result);
			return -1 ;
		}

		/* チェックサム計算*/
		ReadCheckSum = calc_CheckSum(CHECK_SUM_R, data, readSize);

		size -= readSize;

	} while( size );

	gettimeofday(&e, NULL);
	DBGTRACE("データ読み込み時間 = %ld sec\n", (e.tv_sec - s.tv_sec));

	return 0;
}

/* -------------------------------------------------------------------------------- */
/*   OSD-Flash データ書き込み                                                       */
/*                                                                                  */
/*   IN : fd     = ファイルディスクリプタ                                           */
/*        p_data = データ                                                           */
/*        size   = データサイズ                                                     */
/*   OUT:  0    == 正常終了                                                         */
/*        -1    == 異常終了                                                         */
/* -------------------------------------------------------------------------------- */
static int flash_writeData(int fd, const uint8_t *p_data, uint32_t size)
{
	struct timeval s, e;
	struct   flash_dev_ioctl_params param;
	int      result;
	uint32_t writeSize = FLASH_RW_DATA_LEN;
	uint32_t check_sum = 0;

	gettimeofday(&s, NULL);

	/* Chip Erase */
	param.cmd = FLASH_CHIP_ERASE;
	result = ioctl( fd, _IOWR(FLASH_IOCTL_MAGIC, 0, 0), &param );
	if( result ) {
		DBGTRACE("%s(%d) : %d\n", __func__, __LINE__, result);
		return -1;
	}
	gettimeofday(&e, NULL);
	DBGTRACE("データ消去時間 = %ld sec\n", (e.tv_sec - s.tv_sec));

	gettimeofday(&s, NULL);
	do {
		if( writeSize > size ) {
			writeSize = size ;
		}

		/* データ書き込み */
		result = write( fd, p_data, writeSize );
		if( result ) {
			DBGTRACE("%s(%d) : %d\n", __func__, __LINE__, result) ;
			return -1 ;
		}

		/* チェックサム計算*/
		WriteCheckSum = calc_CheckSum( CHECK_SUM_W, p_data, writeSize );

		p_data += writeSize ;
		size   -= writeSize ;

	} while( size ) ;
	gettimeofday(&e, NULL);
	DBGTRACE("データ書き込み時間 = %ld sec\n", (e.tv_sec - s.tv_sec));

	return 0 ;
}

/* -------------------------------------------------------------------------------- */
/*   OSD-Flash アップデート                                                         */
/*                                                                                  */
/*   IN : p_fileName = ファイル名                                                   */
/*   OUT:  0        == 正常終了                                                     */
/*        -1        == 異常終了                                                     */
/* -------------------------------------------------------------------------------- */
static int flash_update( const char *p_fileName )
{
	uint8_t  *p_fwImage ;
	uint32_t  size;
	int       result, fd ;

	/* ファームウェアイメージ読み込み */
	result = readFile( &p_fileName[0], (void *)&p_fwImage );
	if( result < 0 ) {
		DBGTRACE("**** %s\n", &p_fileName[0]) ;
		return -1 ;
	}
	size = (uint32_t)result;

	if( size > (0x00FFFFF0 - 1024) ) {
		DBGTRACE("書き込み容量オーバー : %d\n", size) ;
		free( p_fwImage );
		return -1;
	}

	/* 初期化 */
	WriteCheckSum     = 0;
	WriteEvenCheckSum = 0;
	WriteOddCheckSum  = 0;
	ReadCheckSum      = 0;
	ReadEvenCheckSum  = 0;
	ReadOddCheckSum   = 0;

	do {
		/* 開く */
		fd = open( "/dev/" FLASH_DRV_DEVNAME, O_RDWR );
		if( fd == -1 ) {
			DBGTRACE("can't open : /dev/" FLASH_DRV_DEVNAME "\n");
			break;
		}

		/* データ書き込み */
		DBGTRACE("データ書き込み size = %d\n", size);
		result = flash_writeData( fd, &p_fwImage[0], size );
		if( result < 0 ) {
			DBGTRACE("%s(%d) : %d\n", __func__, __LINE__, result);
			break;
		}

		/* データ読み込み */
		result = flash_read(fd, size);
		if( result < 0 ) {
			DBGTRACE("%s(%d) : %d\n", __func__, __LINE__, result);
			break;
		}

		/* チェックサム値確認 */
		result = check_CheckSum();
		if( result < 0 ) {
			DBGTRACE("%s(%d) : %d\n", __func__, __LINE__, result);
			break;
		}

		/* 正常終了 */
		result = 0;
	} while( 0 );

	if( fd != -1 ) {
		msleep( 100 );
		close( fd );
	}

	free( p_fwImage );

	return result;
}

/* -------------------------------------------------------------------------------- */
/*   メイン処理                                                                     */
/*                                                                                  */
/*   IN : argc = 引数の数                                                           */
/*        argv = 引数                                                               */
/*   OUT: 0   == 正常終了                                                           */
/*        1   == 異常終了                                                           */
/*        2   == ヘルプ表示                                                         */
/* -------------------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
	struct timeval  s, e;
	uint8_t        *p_data ;
	int             result ;

	if( argc == 1 ) {
		fprintf(stderr, "usage: update fwimage\n");
		return 2;
	}

	gettimeofday(&s, NULL);

	result = flash_update( argv[1] );

	gettimeofday(&e, NULL);
	DBGTRACE("総時間 = %ld sec\n", (e.tv_sec - s.tv_sec));

	return 0 ;
}
